import datetime
import os
import sys
import threading
from typing import List, Optional, Type, TypeVar

from pytz import utc
from sqlalchemy import (
    Column,
    DateTime,
    Float,
    LargeBinary,
    SmallInteger,
    String,
    create_engine,
    distinct,
    inspect,
    or_,
    text,
)
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.sql import func
from sqlalchemy.sql.functions import coalesce, count, max, min

from remoulade import Encoder
from remoulade.state import State, StateBackend

Base = declarative_base()

DEFAULT_POSTGRES_URI = "postgresql://remoulade@localhost:5432/remoulade"
DB_VERSION = 3
T = TypeVar("T", bound="StoredState")


class StoredState(Base):

    __tablename__ = "states"

    message_id = Column(String(length=36), primary_key=True, index=True)
    status = Column(String(length=10), index=True)
    actor_name = Column(String(length=79), index=True)
    args = Column(LargeBinary)
    kwargs = Column(LargeBinary)
    options = Column(LargeBinary)
    priority = Column(SmallInteger)
    progress = Column(Float)
    enqueued_datetime = Column(DateTime(timezone=True), index=True)
    started_datetime = Column(DateTime(timezone=True), index=True)
    end_datetime = Column(DateTime(timezone=True), index=True)
    queue_name = Column(String(length=60))
    composition_id = Column(String)

    def as_state(self, encoder: Encoder) -> State:
        state_dict = {}
        mapper = inspect(StoredState)
        for column in mapper.attrs:
            column_value = getattr(self, column.key)
            if column_value is None:
                continue
            if column.key in ["args", "kwargs", "options"]:
                column_value = encoder.decode(column_value)
            state_dict[column.key] = column_value
        return State.from_dict(state_dict)

    @classmethod
    def from_state(cls: Type[T], state: State, max_size: int, encoder: Encoder) -> T:
        state_dict = state.as_dict()
        for key in ["args", "kwargs", "options"]:
            if key in state_dict:
                encoded_value = encoder.encode(state_dict[key])
                state_dict[key] = encoded_value if sys.getsizeof(encoded_value) <= max_size else None
        return cls(**state_dict)


class StateVersion(Base):

    __tablename__ = "version"

    version = Column(SmallInteger, primary_key=True)


def filter_query(
    *,
    query,
    selected_actors: Optional[List[str]],
    selected_statuses: Optional[List[str]],
    selected_message_ids: Optional[List[str]],
    selected_composition_ids: Optional[List[str]],
    start_datetime: Optional[datetime.datetime],
    end_datetime: Optional[datetime.datetime],
):
    if selected_actors is not None:
        query = query.filter(StoredState.actor_name.in_(selected_actors))
    if selected_statuses is not None:
        query = query.filter(StoredState.status.in_(selected_statuses))
    if selected_message_ids is not None:
        query = query.filter(StoredState.message_id.in_(selected_message_ids))
    if selected_composition_ids is not None:
        query = query.filter(StoredState.composition_id.in_(selected_composition_ids))
    if start_datetime is not None:
        query = query.filter(StoredState.enqueued_datetime >= start_datetime)
    if end_datetime is not None:
        query = query.filter(StoredState.enqueued_datetime <= end_datetime)

    return query


class PostgresBackend(StateBackend):
    def __init__(
        self,
        *,
        namespace: str = "remoulade-state",
        encoder: Optional[Encoder] = None,
        client: Optional[sessionmaker] = None,
        max_size: int = 2000000,
        url: Optional[str] = None,
        future: bool = False,
    ):
        self.url = url or os.getenv("REMOULADE_POSTGRESQL_URL") or DEFAULT_POSTGRES_URI
        super().__init__(namespace=namespace, encoder=encoder)
        self.client = client or sessionmaker(create_engine(self.url, pool_pre_ping=True, future=future))
        self.init_db()
        self.max_size = max_size
        self.lock = threading.Lock()

    def init_db(self):
        with self.client.begin() as session:
            bind = session.get_bind()
            insp = inspect(bind)

            if not insp.has_table("version"):
                Base.metadata.create_all(bind=bind, tables=[StateVersion.__table__])

            state_version = session.query(StateVersion).first()
            if state_version is None:
                session.add(StateVersion(version=DB_VERSION))

            if not insp.has_table("states"):
                Base.metadata.create_all(bind=bind, tables=[StoredState.__table__])
            elif state_version is None or state_version.version != DB_VERSION:
                StoredState.__table__.drop(bind=bind)
                Base.metadata.create_all(bind=bind, tables=[StoredState.__table__])
                if state_version is not None:
                    state_version.version = DB_VERSION

    def get_state(self, message_id: str):
        with self.client.begin() as session:
            state = session.query(StoredState).filter_by(message_id=message_id).first()
            if state is None:
                return None
            return state.as_state(self.encoder)

    def set_state(self, state: State, ttl=3600):
        with self.lock:
            with self.client.begin() as session:
                session.merge(StoredState.from_state(state, self.max_size, self.encoder))

    def get_states(
        self,
        *,
        size: Optional[int] = None,
        offset: int = 0,
        selected_actors: Optional[List[str]] = None,
        selected_statuses: Optional[List[str]] = None,
        selected_message_ids: Optional[List[str]] = None,
        selected_composition_ids: Optional[List[str]] = None,
        start_datetime: Optional[datetime.datetime] = None,
        end_datetime: Optional[datetime.datetime] = None,
        sort_column: Optional[str] = None,
        sort_direction: Optional[str] = None,
    ):
        sort_column = sort_column or "enqueued_datetime"
        sort_direction = sort_direction or "desc"

        with self.client.begin() as session:
            query = session.query(StoredState)
            query = filter_query(
                query=query,
                selected_actors=selected_actors,
                selected_statuses=selected_statuses,
                selected_message_ids=selected_message_ids,
                selected_composition_ids=selected_composition_ids,
                start_datetime=start_datetime,
                end_datetime=end_datetime,
            )
            if size is not None:
                query = query.subquery()
                query_group = (
                    session.query(
                        max(query.c.composition_id).label("grouped_composition_id"),
                        max(query.c.message_id).label("grouped_message_id"),
                        max(query.c.status).label("grouped_status"),
                        max(query.c.actor_name).label("grouped_actor_name"),
                        max(query.c.priority).label("grouped_priority"),
                        func.avg(query.c.progress).label("grouped_progress"),
                        min(query.c.enqueued_datetime).label("grouped_enqueued_datetime"),
                        min(query.c.started_datetime).label("grouped_started_datetime"),
                        max(query.c.end_datetime).label("grouped_end_datetime"),
                        max(query.c.queue_name).label("grouped_queue_name"),
                    )
                    .group_by(coalesce(query.c.composition_id, query.c.message_id))
                    .order_by(text(f"grouped_{sort_column} {sort_direction}"))
                )
                query_group = query_group.offset(offset).limit(size).subquery()
                query = (
                    session.query(StoredState)
                    .select_from(StoredState)
                    .join(
                        query_group,
                        or_(
                            StoredState.message_id == query_group.c.grouped_message_id,
                            StoredState.composition_id == query_group.c.grouped_composition_id,
                        ),
                    )
                )
            query = query.order_by(text(f"{sort_column} {sort_direction}"))
            return [state_model.as_state(self.encoder) for state_model in query]

    def get_states_count(
        self,
        *,
        selected_actors: Optional[List[str]] = None,
        selected_statuses: Optional[List[str]] = None,
        selected_messages_ids: Optional[List[str]] = None,
        selected_composition_ids: Optional[List[str]] = None,
        start_datetime: Optional[datetime.datetime] = None,
        end_datetime: Optional[datetime.datetime] = None,
        **kwargs,
    ):
        with self.client.begin() as session:
            query = session.query(count(distinct(coalesce(StoredState.composition_id, StoredState.message_id))))

            query = filter_query(
                query=query,
                selected_actors=selected_actors,
                selected_statuses=selected_statuses,
                selected_message_ids=selected_messages_ids,
                selected_composition_ids=selected_composition_ids,
                start_datetime=start_datetime,
                end_datetime=end_datetime,
            )
            return query.first()[0]

    def clean(self, max_age: Optional[int] = None, not_started: bool = False):
        with self.client.begin() as session:
            query = session.query(StoredState)
            if max_age:
                now = datetime.datetime.now(utc)
                min_datetime = now - datetime.timedelta(minutes=max_age)
                query = session.query(StoredState).filter(StoredState.end_datetime < min_datetime)
            if not_started:
                query = session.query(StoredState).filter(StoredState.started_datetime.is_(None))
            query.delete()



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from contextlib import contextmanager
from queue import Queue
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Set, Type, TypeVar, Union

from .cancel import Cancel, CancelBackend
from .errors import ActorNotFound, NoCancelBackend, NoResultBackend, NoStateBackend
from .logging import get_logger
from .middleware import (
    AgeLimit,
    CatchError,
    CurrentMessage,
    LoggingMetadata,
    MaxMemory,
    MaxTasks,
    MiddlewareError,
    Pipelines,
    Retries,
    ShutdownNotifications,
    TimeLimit,
    default_middleware,
)
from .results import ResultBackend, Results
from .state import MessageState, StateBackend

if TYPE_CHECKING:
    from .actor import Actor
    from .message import Message  # noqa
    from .middleware import Middleware

    M = TypeVar("M", bound=Middleware)
#: The global broker instance.
global_broker: "Optional[Broker]" = None

#: The order in which middlewares are sorted
middleware_order = [
    AgeLimit,
    TimeLimit,
    ShutdownNotifications,
    Pipelines,
    Results,
    CatchError,
    Retries,
    CurrentMessage,
    LoggingMetadata,
    MaxMemory,
    MaxTasks,
    MessageState,
    Cancel,
]

#: A list of extra middleware that will be added to every new instance of Broker
extra_default_middleware: "List[Middleware]" = []


def get_broker() -> "Broker":
    """Get the global broker instance.  If no global broker is set,
    this initializes a RabbitmqBroker and returns it.

    Returns:
      Broker: The default Broker.
    Raises:
      ValueError if no broker is set
    """
    global global_broker
    if global_broker is None:
        raise ValueError("Broker not found, are you sure you called set_broker(broker) ?")
    return global_broker


def set_broker(broker: "Broker") -> None:
    """Configure the global broker instance.

    Parameters:
      broker(Broker): The broker instance to use by default.
    """
    global global_broker
    global_broker = broker


def change_broker(broker: "Broker") -> None:
    """Change of broker instance

    It will take all the actors of the current broker and add them to the new one

    Parameters:
      broker(Broker): The broker instance to use by default.
    """
    global global_broker
    actors = global_broker.actors.values() if global_broker else []
    global_broker = broker
    declare_actors(actors)


def add_extra_default_middleware(middleware: "Middleware") -> None:
    """Configure an extra middleware instance that will be added by default to all new Broker instances.
    It will also be added to the current global_broker, if it exists.

    Parameters:
      middleware(Middleware): The middleware to add.
    """
    global extra_default_middleware
    extra_default_middleware.append(middleware)

    try:
        get_broker().add_middleware(middleware)
    except ValueError:
        pass


def remove_extra_default_middleware(middleware_class: "Type[Middleware]") -> None:
    """Remove an extra default middleware so that it won't be added to new Broker instances.
    It will also be removed from the current global_broker, if it exists.

    Parameters:
        middleware_class(Type[Middleware]): The middleware class.
    """
    global extra_default_middleware
    extra_default_middleware = [m for m in extra_default_middleware if not isinstance(m, middleware_class)]

    try:
        get_broker().remove_middleware(middleware_class)
    except ValueError:
        pass


def declare_actors(actors: "Iterable[Actor]") -> None:
    """Declare the given actors to the current broker

    Parameters:
      actors(List[Actor]): The actors being declared.
    """
    broker = get_broker()
    for actor in actors:
        broker.declare_actor(actor)


class Broker:
    """Base class for broker implementations.

    Parameters:
      middleware(list[Middleware]): The set of middleware that apply
        to this broker.  If you supply this parameter, you are
        expected to declare *all* middleware.  Most of the time,
        you'll want to use :meth:`.add_middleware` instead.

    Attributes:
      actor_options(set[str]): The names of all the options actors may
        overwrite when they are declared.
    """

    def __init__(self, middleware: "Optional[Iterable[Middleware]]" = None):
        self.logger = get_logger(__name__, type(self))
        self.actors: "Dict[str, Actor]" = {}
        self.queues: Dict[str, Optional[Queue]] = {}
        self.delay_queues: Set[str] = set()

        self.actor_options: Set[str] = set()
        self.middleware: "List[Middleware]" = []
        self.group_transaction = False

        if middleware is None:
            middleware = [m() for m in default_middleware]

        for m in middleware:
            self.add_middleware(m)

        for m in extra_default_middleware:
            self.add_middleware(m)

    @property
    def local(self) -> bool:
        return False

    @contextmanager
    def tx(self):
        yield

    def emit_before(self, signal, *args, **kwargs):
        for middleware in self.middleware:
            try:
                getattr(middleware, "before_" + signal)(self, *args, **kwargs)
            except MiddlewareError:
                raise
            except Exception:
                self.logger.critical("Unexpected failure in before_%s.", signal, exc_info=True)

    def emit_after(self, signal, *args, **kwargs):
        for middleware in reversed(self.middleware):
            try:
                getattr(middleware, "after_" + signal)(self, *args, **kwargs)
            except Exception:
                self.logger.critical("Unexpected failure in after_%s.", signal, exc_info=True)

    def get_result_backend(self) -> ResultBackend:
        """Get the ResultBackend associated with the broker

        Raises:
            NoResultBackend: if there is no ResultBackend

        Returns:
            ResultBackend: the result backend
        """
        return self._get_backend("results")

    def get_cancel_backend(self) -> CancelBackend:
        """Get the CancelBackend associated with the broker

        Raises:
            NoCancelBackend: if there is no CancelBackend

        Returns:
            CancelBackend: the cancel backend
        """
        return self._get_backend("cancel")

    def get_state_backend(self) -> StateBackend:
        """Get the StateBackend associated with the broker

        Raises:
            NoStateBackend: if there is no StateBackend

        Returns:
            StateBackend: the state backend
        """
        return self._get_backend("state")

    def _get_backend(self, name: str):
        """Get the backend associated with the broker either cancel or results"""
        message = "The default broker doesn't have a %s backend."
        backends = {
            "results": (Results, NoResultBackend(message % "results")),
            "cancel": (Cancel, NoCancelBackend(message % "cancel")),
            "state": (MessageState, NoStateBackend(message % "state")),
        }
        try:
            middleware_class, exception = backends[name]
            middleware: Optional[Union[Results, Cancel, MessageState]] = self.get_middleware(middleware_class)
            if middleware is not None:
                return middleware.backend
            else:
                raise exception
        except KeyError as e:
            raise ValueError("invalid backend name") from e

    def add_middleware(self, middleware: "Middleware") -> None:
        """Add a middleware object to this broker.  The middleware is
         added to his default position.


        Parameters:
          middleware(Middleware): The middleware.
        """
        middleware_order_count = 0
        added_middleware_count = 0

        while added_middleware_count < len(self.middleware) and middleware_order_count < len(middleware_order):
            current_middleware = middleware_order[middleware_order_count]
            if isinstance(middleware, current_middleware):
                if isinstance(self.middleware[added_middleware_count], current_middleware):
                    self.middleware[added_middleware_count] = middleware
                else:
                    self.middleware.insert(added_middleware_count, middleware)
                break
            if isinstance(self.middleware[added_middleware_count], current_middleware):
                added_middleware_count += 1
            middleware_order_count += 1
        else:
            self.middleware.append(middleware)

        self.actor_options |= middleware.actor_options

        for actor_name in self.get_declared_actors():
            middleware.after_declare_actor(self, actor_name)

        for queue_name in self.get_declared_queues():
            middleware.after_declare_queue(self, queue_name)

        for queue_name in self.get_declared_delay_queues():
            middleware.after_declare_delay_queue(self, queue_name)

    def get_middleware(self, middleware_class: "Type[M]") -> "Optional[M]":
        for middleware in self.middleware:
            if isinstance(middleware, middleware_class):
                return middleware
        return None

    def remove_middleware(self, middleware_class: "Type[Middleware]"):
        """Removes a middleware object from this broker.


        Parameters:
          middleware_class(Type[Middleware]): The middleware class.
        """
        self.middleware = [m for m in self.middleware if not isinstance(m, middleware_class)]

    def close(self):
        """Close this broker and perform any necessary cleanup actions."""

    def consume(self, queue_name: str, prefetch: int = 1, timeout: int = 30000) -> "Consumer":  # pragma: no cover
        """Get an iterator that consumes messages off of the queue.

        Raises:
          QueueNotFound: If the given queue was never declared.

        Parameters:
          queue_name(str): The name of the queue to consume messages off of.
          prefetch(int): The number of messages to prefetch per consumer.
          timeout(int): The amount of time in milliseconds to idle for.

        Returns:
          Consumer: A message iterator.
        """
        raise NotImplementedError

    def declare_actor(self, actor: "Actor") -> None:  # pragma: no cover
        """Declare a new actor on this broker.  Declaring an Actor
        twice replaces the first actor with the second by name.

        Parameters:
          actor(Actor): The actor being declared.
        """
        actor.set_broker(self)
        self.emit_before("declare_actor", actor)
        for queue_name in actor.queue_names:
            self.declare_queue(queue_name)
        self.actors[actor.actor_name] = actor
        self.emit_after("declare_actor", actor)

    def declare_queue(self, queue_name: str) -> None:  # pragma: no cover
        """Declare a queue on this broker.  This method must be
        idempotent.

        Parameters:
          queue_name(str): The name of the queue being declared.
        """
        raise NotImplementedError

    def _apply_delay(self, message: "Message", delay: Optional[int] = None) -> "Message":
        raise NotImplementedError

    def _enqueue(self, message: "Message", *, delay: Optional[int] = None) -> "Message":
        raise NotImplementedError

    def enqueue(self, message: "Message", *, delay: Optional[int] = None) -> "Message":  # pragma: no cover
        """Enqueue a message on this broker.

        Parameters:
          message(Message): The message to enqueue.
          delay(int): The number of milliseconds to delay the message for.

        Returns:
          Message: Either the original message or a copy of it.
        """

        message = self._apply_delay(message, delay)
        self.emit_before("enqueue", message, delay)

        try:
            message = self._enqueue(message, delay=delay)
            self.emit_after("enqueue", message, delay)
            return message

        except BaseException as e:
            self.emit_after("enqueue", message, delay, exception=e)
            raise e from e

    def get_actor(self, actor_name: str) -> "Actor":  # pragma: no cover
        """Look up an actor by its name.

        Parameters:
          actor_name(str): The name to look up.

        Raises:
          ActorNotFound: If the actor was never declared.

        Returns:
          Actor: The actor.
        """
        try:
            return self.actors[actor_name]
        except KeyError:
            raise ActorNotFound(actor_name) from None

    def get_declared_actors(self) -> Set[str]:  # pragma: no cover
        """Get all declared actors.

        Returns:
          set[str]: The names of all the actors declared so far on
          this Broker.
        """
        return set(self.actors.keys())

    def get_declared_queues(self) -> Set[str]:  # pragma: no cover
        """Get all declared queues.

        Returns:
          set[str]: The names of all the queues declared so far on
          this Broker.
        """
        return set(self.queues.keys())

    def get_declared_delay_queues(self) -> Set[str]:  # pragma: no cover
        """Get all declared delay queues.

        Returns:
          set[str]: The names of all the delay queues declared so far
          on this Broker.
        """
        return self.delay_queues.copy()

    def flush(self, queue_name: str) -> None:  # pragma: no cover
        """Drop all the messages from a queue.

        Parameters:
          queue_name(str): The name of the queue to flush.
        """
        raise NotImplementedError()

    def flush_all(self) -> None:  # pragma: no cover
        """Drop all messages from all declared queues."""
        raise NotImplementedError()

    def join(self, queue_name: str, *, timeout: Optional[int] = None) -> None:  # pragma: no cover
        """Wait for all the messages on the given queue to be processed.
        This method is only meant to be used in tests to wait for all the messages in a queue to be processed."""
        raise NotImplementedError()


class Consumer:
    """Consumers iterate over messages on a queue.

    Consumers and their MessageProxies are *not* thread-safe.
    """

    def __iter__(self):  # pragma: no cover
        """Returns this instance as a Message iterator."""
        return self

    def ack(self, message):  # pragma: no cover
        """Acknowledge that a message has been processed, removing it
        from the broker.

        Parameters:
          message(MessageProxy): The message to acknowledge.
        """
        raise NotImplementedError

    def nack(self, message):  # pragma: no cover
        """Move a message to the dead-letter queue.

        Parameters:
          message(MessageProxy): The message to reject.
        """
        raise NotImplementedError

    def requeue(self, messages):  # pragma: no cover
        """Move unacked messages back to their queues.  This is called
        by consumer threads when they fail or are shut down.  The
        default implementation does nothing.

        Parameters:
          messages(list[MessageProxy]): The messages to requeue.
        """

    def __next__(self):  # pragma: no cover
        """Retrieve the next message off of the queue.  This method
        blocks until a message becomes available.

        Returns:
          MessageProxy: A transparent proxy around a Message that can
          be used to acknowledge or reject it once it's done being
          processed.
        """
        raise NotImplementedError

    def close(self):
        """Close this consumer and perform any necessary cleanup actions."""


class MessageProxy:
    """Base class for messages returned by :meth:`Broker.consume`."""

    def __init__(self, message):
        self.failed = False
        self._message = message

    def fail(self):
        """Mark this message for rejection."""
        self.failed = True

    def __getattr__(self, name):
        return getattr(self._message, name)

    def __str__(self):
        return str(self._message)

    def __lt__(self, other):
        # This can get called if two messages have the same priority
        # in a queue.  If that's the case, we don't care which runs
        # first.
        return True

    def __eq__(self, other):
        if isinstance(other, MessageProxy):
            return self._message == other._message
        return self._message == other



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import traceback
from typing import Callable, Optional

from ..helpers import compute_backoff
from ..helpers.backoff import BackoffStrategy
from ..logging import get_logger
from .middleware import Middleware

#: The default minimum amount of backoff to apply to retried tasks.
DEFAULT_MIN_BACKOFF = 15000

#: The default maximum amount of backoff to apply to retried tasks.
DEFAULT_MAX_BACKOFF = 1000 * 60 * 60


class Retries(Middleware):
    """Middleware that automatically retries failed tasks with
    exponential backoff.

    Parameters:
      max_retries(int): The maximum number of times tasks can be retried.
      min_backoff(int): The minimum amount of backoff milliseconds to
        apply to retried tasks.  Defaults to 15 seconds.
      max_backoff(int): The maximum amount of backoff milliseconds to
        apply to retried tasks.  Defaults to 7 days.
      retry_when(Callable[[int, Exception], bool]): An optional
        predicate that can be used to programmatically determine
        whether a task should be retried or not.  This takes
        precedence over `max_retries` when set.
      increase_priority_on_retry(bool): specifies wether to increase
        priority of message when retried. Default is False.
    """

    def __init__(
        self,
        *,
        max_retries: Optional[int] = None,
        min_backoff: Optional[int] = None,
        max_backoff: Optional[int] = None,
        retry_when: Optional[Callable[[int, Exception], bool]] = None,
        backoff_strategy: BackoffStrategy = "exponential",
        jitter: bool = True,
        increase_priority_on_retry: bool = False
    ):
        self.logger = get_logger(__name__, type(self))
        self.max_retries = max_retries
        self.min_backoff = min_backoff or DEFAULT_MIN_BACKOFF
        self.max_backoff = max_backoff or DEFAULT_MAX_BACKOFF
        self.retry_when = retry_when
        self.backoff_strategy = backoff_strategy
        self.jitter = jitter
        self.increase_priority_on_retry = increase_priority_on_retry

    @property
    def actor_options(self):
        return {
            "max_retries",
            "min_backoff",
            "max_backoff",
            "retry_when",
            "backoff_strategy",
            "jitter",
            "increase_priority_on_retry",
        }

    def after_process_message(self, broker, message, *, result=None, exception=None):
        if exception is None:
            return

        retries = message.options.setdefault("retries", 0)
        max_retries = self.get_option("max_retries", broker=broker, message=message)
        retry_when = self.get_option("retry_when", broker=broker, message=message)
        if retry_when is None and not max_retries:
            message.fail()
            return

        if (
            retry_when is not None
            and not retry_when(retries, exception)
            or max_retries is not None
            and retries >= max_retries
        ):
            if max_retries is not None and retries >= max_retries:
                self.logger.warning(f"Retries exceeded for message {message.message_id}.")
            else:
                self.logger.warning(f"Message {message.message_id} has failed and will not be retried.")
            message.fail()
            return

        new_message = message.copy()
        increase_priority_on_retry = self.get_option("increase_priority_on_retry", broker=broker, message=message)
        if increase_priority_on_retry and getattr(broker, "max_priority", None) is not None:
            new_message.options["priority"] = min(message.options.get("priority", 0) + 1, broker.max_priority)
            new_message.options["increase_priority_on_retry"] = False  # we only want to do it once

        new_message.options["retries"] += 1
        new_message.options["traceback"] = traceback.format_exc(limit=30)
        min_backoff = self.get_option("min_backoff", broker=broker, message=message)
        max_backoff = self.get_option("max_backoff", broker=broker, message=message)
        backoff_strategy = self.get_option("backoff_strategy", broker=broker, message=message)
        jitter = self.get_option("jitter", broker=broker, message=message)
        _, backoff = compute_backoff(
            retries,
            min_backoff=min_backoff,
            max_backoff=max_backoff,
            jitter=jitter,
            max_retries=max_retries or 10,
            backoff_strategy=backoff_strategy,
        )
        self.logger.info("Retrying message %r in %d milliseconds.", message.message_id, backoff)
        broker.enqueue(new_message, delay=backoff)



# This file is a part of Remoulade.
#
# Copyright (C) 2017,2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# Remoulade is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Remoulade is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import uuid
from collections.abc import Iterable
from itertools import islice
from time import time
from typing import Iterator, List


def current_millis():
    """Returns the current UNIX time in milliseconds."""
    return int(time() * 1000)


def generate_unique_id() -> str:
    """Generate a globally-unique message id."""
    return str(uuid.uuid4())


def flatten(iterable):
    """Flatten deep an iterable"""
    for el in iterable:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


def chunk(iterable: Iterable, size: int) -> Iterator[List]:
    """Returns an iterator of a list of length size"""
    i = iter(iterable)
    piece = list(islice(i, size))
    while piece:
        yield piece
        piece = list(islice(i, size))
